"""
The AwsIamTester class implements all necessary logic to run validations on an account, role or user.
"""

# pylint: disable=broad-except,C0103,E0401,R0912,R0913,R0914,R0915,R1702,W0603,W1203

from __future__ import annotations

import os
import sys
import errno
import json
import logging
import re
import time
import yaml
import click
import boto3 # type: ignore
import botocore # type: ignore

from tabulate import tabulate
from typing import Any, Dict, List, Optional, Tuple, Union  #, Literal # Literal is p3.8 and higher
from termcolor import colored

class AwsIamTester():
    # Defaults
    DEFAULT_SLEEP_SECONDS = 300

    logger: logging.Logger
    logger_initialized: bool

    # define boto3 retry logic, as the simulation api might do some throttling
    boto3_config = botocore.config.Config(
        retries=dict(
            max_attempts=10
        )
    )

    def __init__(
        self,
        debug: bool,
        ):

        self.logger_initialized = False
        self.debug = debug

    def get_aws_data(self):
        logger = self.get_logger()

        account_id = None
        account_alias = None

        # first get current account id
        sts_client = boto3.client("sts")
        account_id = sts_client.get_caller_identity()["Account"]

        # try to get the friendly account name
        iam_client = boto3.client("iam")
        try:
            account_alias = iam_client.list_account_aliases(MaxItems=1)["AccountAliases"][0]
        except (KeyError, IndexError):
            account_alias = account_id
        
        return account_id, account_alias

    def check_action(
        self,
        user: str,
        role: str,
        action: str,
        resource: str,
        json_output: bool,
    ):
        "Checks whether a given user OR role has access to a particular action and resource"
        try:
            # logger = self.get_logger()
            account_id, account_alias = self.get_aws_data()

            if user and role:
                raise Exception("Pass in user or role, not both")
            elif user:
                source = f"arn:aws:iam::{account_id}:user/{user}"
            elif role:
                source = f"arn:aws:iam::{account_id}:role/{role}"
            else:
                raise Exception("One of user or role is required")

            results, counter = self.evaluate_sources(
                sources=[source],
                limit_to=[],
                exemptions=[],
                expect_failures=None, # this will then both denied and allows
                actions=[action],
                resources=[resource],
                sim_context=[],
            )
            if json_output:
                return self.handle_results(
                    results=results,
                )
            else:
                result = results[0]

                decision = result['decision']
                if decision == "allowed":
                    allowed = True
                    colour = "green"
                else:
                    allowed = False
                    colour = "red"

                org_scp = result['org_scp']
                if org_scp == "allowed":
                    org_colour = "green"
                elif org_scp == "denied":
                    org_colour = "red"
                else:
                    org_colour = "white"

                click.secho(f"\n\nTest:", bold=True)
                click.echo(f"Source:            {source}")
                click.echo(f"Action:            {action}")
                click.echo(f"Resource:          {resource}")
                click.secho(f"Result:            ", nl=False)
                click.secho(f"{decision}", fg=colour)
                click.secho(f"Allowed by org:    ", nl=False)
                click.secho(f"{org_scp}\n", fg=org_colour)

                ms_key = "matched_statements"
                if ms_key in result:
                    matched_statements = result[ms_key]

                    click.secho("Matched statements:", bold=True)
                    for ms in matched_statements:
                        click.echo(f"Policy:     {ms['SourcePolicyId']}")
                        click.echo(f"Type:       {ms['SourcePolicyType']}")
                        click.echo(f"Start:      L{ms['StartPosition']['Line']}:C{ms['StartPosition']['Column']}")
                        click.echo(f"End:        L{ms['EndPosition']['Line']}:C{ms['EndPosition']['Column']}")
                
                return allowed
        except botocore.exceptions.ClientError as ce:
            if re.match("^(.*)(security token|AccessDenied)(.*)$", str(ce)):
                click.echo(f"Please make sure you are logged in into AWS, with sufficient permissions.")

            raise

    def check_access(
        self,
        action: str,
        resource: str,
        json_output: bool,
    ):
        try:
            logger = self.get_logger()
            account_id, account_alias = self.get_aws_data()

            sources = self.determine_source(
                account_id=account_id,
                no_system_roles=False,
            )
            results, counter = self.evaluate_sources(
                sources=sources,
                limit_to=[],
                exemptions=[],
                expect_failures=True, # this will then return people with access
                actions=[action],
                resources=[resource],
                sim_context=[],
            )

            if json_output:
                return self.handle_results(
                    results=results,
                )
            else:
                to_print = []
                for r in results:
                    policies = ""
                    ms_key = "matched_statements"
                    if ms_key in r:
                        matched_statements = r[ms_key]

                        for ms in matched_statements:
                            policies = policies + ", " + ms['SourcePolicyId']

                        policies = policies[2:] # remove the first ', '

                    to_print.append([
                        r['source'],
                        r['action'],
                        r['resource'],
                        r['decision'],
                        policies,
                    ])

                self.show_summary(to_print)

        except botocore.exceptions.ClientError as ce:
            if re.match("^(.*)(security token|AccessDenied)(.*)$", str(ce)):
                click.echo(f"Please make sure you are logged in into AWS, with sufficient permissions.")

            raise

    def check_account(
        self, 
        number_of_runs: int,
        dry_run: bool,
        config_file: str,
        no_system_roles: bool,
        write_to_file: bool,
        output_location: str,
    ) -> int:
        """
        Run the IAM policy tests. It returns a dict with findings (or empty if none)
        """
        def convert_context_dict(my_dict: Dict) -> Dict:
            result = {}
            for key, value in my_dict.items():
                # convert key to CamelCase
                if isinstance(key, str):
                    new_key = ''.join(x.capitalize() or '_' for x in key.split('_'))
                    result[new_key] = value

                    # the ContextKeyValues should be a list
                    if new_key == 'ContextKeyValues':
                        result[new_key] = [value]
                else:
                    result[key] = value

            # convert boolean values to strings
            if result['ContextKeyType'] == 'boolean':
                result['ContextKeyValues'] = [str(x).lower() for x in result['ContextKeyValues']]

            return result

        try:
            logger = self.get_logger()
            account_id, account_alias = self.get_aws_data()

            config, global_exemptions, global_limit_to, user_landing_account = self.read_config(config_file)

            sources = self.determine_source(
                user_landing_account=user_landing_account,
                account_id=account_id,
                no_system_roles=no_system_roles,
                global_limit_to=global_limit_to,
                global_exemptions=global_exemptions,
            )
            results = []

            counter = 0
            logger.debug("Start checking the configs")
            for cfg in config['tests']:
                # do we need to break?
                if 0 < number_of_runs < counter:
                    break

                actions = cfg["actions"]
                resources = [x.format(account_id=account_id) for x in cfg["resources"]]

                # create a full list of exemptions
                exemptions = []
                if "exemptions" in cfg:
                    exemptions = cfg["exemptions"]

                # check if this test contain a 'limit_to' element
                limit_to = []
                if "limit_to" in cfg:
                    limit_to = cfg["limit_to"]

                # check if the expected result has the expected values
                expected_result = cfg["expected_result"].lower()
                if expected_result not in ['fail', 'succeed']:
                    raise ValueError(
                        f"The expected result should be 'fail' or 'succeed'. It is now '{expected_result}'"
                    )

                expect_failures = (expected_result == "fail")
                logger.debug(f"Run config for actions: {actions} with resources {resources}")

                # check if we have a custom context
                sim_context = None
                if "custom_context" in cfg:
                    custom_context = cfg["custom_context"]
                    # the keys in the custom context need to be converted to CamelCase
                    sim_context = []
                    for ctx in custom_context:
                        c_ctx = convert_context_dict(ctx)
                        sim_context.append(c_ctx)

                source_results, counter = self.evaluate_sources(
                    sources=sources,
                    limit_to=limit_to,
                    exemptions=exemptions,
                    number_of_runs=number_of_runs,
                    dry_run=dry_run,
                    expect_failures=expect_failures,
                    counter=counter,
                    actions=actions,
                    resources=resources,
                    sim_context=sim_context,
                )
                results.extend(source_results)

            return self.handle_results(
                results=results,
                write_to_file=write_to_file,
                output_location=output_location,
                account=account_alias,
            )
        except botocore.exceptions.ClientError as ce:
            if re.match("^(.*)(security token|AccessDenied)(.*)$", str(ce)):
                click.echo(f"Please make sure you are logged in into AWS, with sufficient permissions.")

            raise

    def get_logger(self) -> logging.Logger:
        "Set up the logger (if not done yet) and returns it"
        if not self.logger_initialized:            
            self.logger = logging.getLogger('iam-tester')
            self.logger.propagate = False
            if not self.logger.handlers:
                ch = logging.StreamHandler(sys.stdout)
                if self.debug:
                    self.logger.setLevel(level=logging.DEBUG)
                    ch.setLevel(level=logging.DEBUG)
                else:
                    self.logger.setLevel(level=logging.INFO)
                    ch.setLevel(level=logging.INFO)
                formatter = logging.Formatter(
                    "%(levelname)s:\t%(message)s"
                )
                ch.setFormatter(formatter)
                self.logger.addHandler(ch)

        return self.logger

    def read_config(self, config_file: str) -> Tuple[Dict, List[str], List[str], Optional[str]]:
        "Read and parse config file"
        logger = self.get_logger()
        logger.debug(f"Read config file {config_file}")

        file = None
        try:
            with open(config_file) as file:
                config = yaml.load(file, Loader=yaml.FullLoader)
        except IOError:
            logger.error(f"Config '{config_file}' not accessible")
            raise
        finally:
            try:
                if file:
                    file.close()
            except IOError:
                pass

        global_exemptions = []
        if "global_exemptions" in config:
            global_exemptions = config["global_exemptions"]

        global_limit_to = []
        if "global_limit_to" in config:
            global_limit_to = config["global_limit_to"]

        user_landing_account = None
        if "user_landing_account" in config:
            user_landing_account = config["user_landing_account"]

        return config, global_exemptions, global_limit_to, user_landing_account

    def get_iam_roles(
            self,
            user_landing_account: Optional[str],
            my_account: str,
            no_system_roles: bool
        ) -> List[str]:
        "Read (and filter) the IAM roles in the account"
        logger = self.get_logger()
        client = boto3.client('iam')
        roles = []

        paginator = client.get_paginator('list_roles')
        page_iterator = paginator.paginate()

        for page in page_iterator:
            rolelist = page["Roles"]
            for role in rolelist:
                # only get the roles that can be assumed by 'users'
                policy_doc = json.dumps(role["AssumeRolePolicyDocument"])

                if 'aws-service-role' not in role["Path"] and ( # ignore service roles
                        # accept roles that can be assumed by users in my account
                        f"arn:aws:iam::{my_account}:root" in policy_doc or
                        # accept roles that can be assumed by a dedicated user landing account
                        f"arn:aws:iam::{user_landing_account}:root" in policy_doc or
                        # accept roles that can be assumed with SAML
                        f"arn:aws:iam::{my_account}:saml-provider" in policy_doc or
                        # do we want to include non user assumable roles
                        not no_system_roles
                    ):
                    roles.append(role['Arn'])

        return roles

    def get_iam_users(self) -> List[str]:
        "Get all IAM users"
        client = boto3.client('iam')
        users = []

        paginator = client.get_paginator('list_users')
        page_iterator = paginator.paginate()
        for page in page_iterator:
            userlist = page["Users"]
            for user in userlist:
                users.append(user['Arn'])

        return users

    def determine_source(
            self,
            account_id: str,
            user_landing_account: Optional[str] = None,
            no_system_roles: bool = False,
            global_limit_to: List[str] = [],
            global_exemptions: List[str] = [],
        ) -> List[str]:
        "Determine the list of sources that needs to be evaluated"
        logger = self.get_logger()

        logger.debug("dynamically collect users and roles")
        users = self.get_iam_users()
        roles = self.get_iam_roles(
            user_landing_account=user_landing_account,
            my_account=account_id,
            no_system_roles=no_system_roles
            )

        sources = []
        sources.extend(users)
        sources.extend(roles)
        len_1 = len(sources)
        
        # If we have a global_limit_to, reduce the full set of sources
        # to only that in order to improve performance
        if global_limit_to:
            filtered_sources = []
            for limit in global_limit_to:
                compiled = re.compile(limit)
                filtered_list = [s for s in sources if compiled.match(s)]
                filtered_sources.extend(filtered_list)
            # Now remove duplicates
            sources = list(set(filtered_sources))
            logger.debug(f"Number of sources limited from {len_1} to {len(sources)}")

        # If we have a global_exemptions, remove these from the sources as well
        if global_exemptions:
            exempt_sources = []
            for limit in global_exemptions:
                compiled = re.compile(limit)
                filtered_list = [s for s in sources if compiled.match(s)]
                exempt_sources.extend(filtered_list)
            # Now remove duplicates and subtract from sources list
            sources = list(set(sources) - set(exempt_sources))
            logger.debug(f"Number of sources reduced with exemptions from {len_1} to {len(sources)}")

        return sources

    def evaluate_sources(
        self,
        sources: List[str],
        limit_to: List[str],
        exemptions: List[str],
        expect_failures: Optional[int],  # This is better but from p3.8 only: Optional[Union[Literal[True], Literal[False]]],
        actions: List[str],
        resources: List[str],
        sim_context: List[Dict[Any, Any]],
        number_of_runs: int = -1,
        dry_run: bool = False,
        counter: int = 0,
        ) -> Tuple[List[Dict], int]:
        "Evaluate the list of sources for a given configuration"

        logger = self.get_logger()
        logger.debug(f"Evaluate sources: {sources}")

        results = []
        for source in sources:
            if limit_to: # do we have a limit_to?
                exempt = True
                for limit in limit_to:
                    if re.match(limit, source):
                        exempt = False
                        break

                if exempt and self.debug:
                    logger.debug(f"\nSource {source} is not included in whitelist")
                elif exempt:
                    click.secho(".", fg="blue", nl=False)
                    sys.stdout.flush()
            else: # or is this source exempt from testing
                exempt = False
                for exemption in exemptions:
                    if re.match(exemption, source):
                        exempt = True
                        if self.debug:
                            logger.debug(
                                f"\nSource {source} is exempt from testing using: {exempt}"
                            )
                        else:
                            click.secho(".", fg="blue", nl=False)
                            sys.stdout.flush()
                        break

            counter += 1
            if 0 < number_of_runs < counter:
                click.echo("\n")
                logger.debug("Test run mode activated, max iterations reached.")
                break
            if dry_run:
                logger.debug(f"Dry run mode, no simulation for {source}")
            elif not exempt:
                evaluation_results = self.simulate_policy(
                    source=source,
                    actions=actions,
                    resources=resources,
                    sim_context=sim_context,
                )

                if evaluation_results:
                    denies = [x for x in evaluation_results if self.is_denied(x)]
                    allows = [x for x in evaluation_results if not self.is_denied(x)]

                    if expect_failures is None:
                        # just use all results
                        filtered_results = evaluation_results
                    elif expect_failures:
                        # allows are failures
                        filtered_results = allows
                    else:
                        # denies are failures
                        filtered_results = denies

                    success = (len(filtered_results) == 0)

                    if expect_failures is None:
                        results.extend(
                            self.construct_results(
                                source=source,
                                results=filtered_results,
                                print_results=self.debug,
                            )
                        )
                        if not self.debug:
                            if filtered_results[0]["EvalDecision"] == "allowed":
                                fg = "green"
                            else:
                                fg = "red"

                            click.secho(".", fg=fg, nl=False)
                            sys.stdout.flush()
                    elif success:
                        if self.debug:
                            click.secho(f"Success: {source}", fg="green")
                        else:
                            click.secho(".", fg="green", nl=False)
                            sys.stdout.flush()
                    else:
                        results.extend(
                            self.construct_results(
                                source=source,
                                results=filtered_results,
                                print_results=self.debug,
                            )
                        )

                        if not self.debug:
                            click.secho(".", fg="red", nl=False)
                            sys.stdout.flush()

        return results, counter

    def simulate_policy(
            self,
            source: str,
            actions: List[str],
            resources: List[str],
            sim_context: List[Dict] = None
            ) -> List[Dict]:
        """Simulate a set of actions from a specific principal against a resource"""
        def simulate(
                source: str,
                actions: List[str],
                resources: List[str],
                sim_context: List[Dict] = None
            ) -> List[Dict]:
            # do we have a custom context to pass along?
            if not sim_context:
                sim_context = [
                    {
                        'ContextKeyName': 'aws:MultiFactorAuthPresent',
                        'ContextKeyValues': [
                            # Always test with MFA present, which is worst case scenario
                            # (as lots of rights might be revoked without MFA)
                            'true',
                        ],
                        'ContextKeyType': 'boolean',
                    },
                    {
                        'ContextKeyName': 'redshift:DbUser',
                        'ContextKeyValues': [
                            'admin',
                        ],
                        'ContextKeyType': 'string',
                    },
                ]

            response = client.simulate_principal_policy(
                PolicySourceArn=source,
                ActionNames=actions,
                ResourceArns=resources,
                ContextEntries=sim_context,
            )
            return response["EvaluationResults"]

        logger = self.get_logger()

        client = boto3.client("iam", config=self.boto3_config)
        try:
            return simulate(source, actions, resources, sim_context)
        except client.exceptions.NoSuchEntityException as nsee:
            click.echo("\n")
            logger.error(
                f"Could not find entity {source} during simulation, has it just been removed?\n{nsee}"
            )
            # but ignore it
            return []
        except client.exceptions.ClientError as ce:
            if "throttling" in str(ce).lower():
                logger.error(
                    colored(
                        "\nThrottling of API is requested. " +
                        f"Sleep for {self.DEFAULT_SLEEP_SECONDS} seconds and try again\n"
                    ),
                    "blue"
                )
                time.sleep(self.DEFAULT_SLEEP_SECONDS)
                return simulate(source, actions, resources, sim_context)
            else:
                raise

    def is_denied(self, evaluationResults: Dict) -> bool:
        "Returns whether or not the evaluation is denied"
        return evaluationResults["EvalDecision"] != "allowed"

    def construct_results(
            self,
            source: str,
            results: Any,
            print_results: bool = True,
        ) -> List[Dict]:
        """Constructs a dict with the results of a simulation evaluation result"""
        output = ""
        response = []

        for er in results:
            if er.get("OrganizationsDecisionDetail", None):
                if er.get("OrganizationsDecisionDetail").get("AllowedByOrganizations"):
                    org_scp = "allowed"
                else:
                    org_scp = "denied"
            else:
                org_scp = "no_org"

            message = (
                f"\nSource: {source}\n"
                f"Evaluated Action Name: {er['EvalActionName']}\n"
                f"\tEvaluated Resource name: {er['EvalResourceName']}\n"
                f"\tDecision: {er['EvalDecision']}\n"
                f"\tOrganizations policy: {org_scp}\n"
                f"\tMatched statements: {er['MatchedStatements']}"
            )
            r = {
                "source": source,
                "action": er['EvalActionName'],
                "resource": er['EvalResourceName'],
                "decision": er['EvalDecision'],
                "org_scp": org_scp,
                "matched_statements": er['MatchedStatements'],
            }
            response.append(r)
            output += message
        if print_results:
            click.secho(output, fg="red")

        return response

    def handle_results(
            self,
            results: List[Dict],
            write_to_file: bool = False,
            output_location: str = "",
            account: str = "",
        ) -> int:
        "Print the results or write them to file"
        logger = self.get_logger()

        click.echo("\n\n")
        logger.debug("Handle results")
        return_value = 1
        if results and write_to_file:
            timestr = time.strftime("%Y%m%d-%H%M%S")
            filename = f'results-{account}-{timestr}.json'

            # remove the trailing / if present
            output_location = output_location.rstrip('/')

            # do we need to write to s3?
            if output_location.startswith('s3://'):
                # parse the output location into bucket and key
                bucket = output_location[5:].split("/")[0]
                # do we have a prefix?
                try:
                    prefix = output_location[5:].split("/", 1)[1]
                except IndexError:
                    prefix = ""

                s3_client = boto3.client('s3')

                full_filename = f'{prefix}/{filename}'
                logger.debug(f"Try to write results to s3://{bucket}/{full_filename}")
                s3_client.put_object(
                    Body=json.dumps(results).encode(),
                    Bucket=bucket,
                    Key=full_filename,
                )
            else:
                # first ensure directory exists
                try:
                    if os.path.isabs(output_location): # is output location absolute?
                        path = output_location
                    else:
                        path = os.path.abspath(output_location)
                    os.makedirs(path)
                except OSError as exc:
                    if exc.errno == errno.EEXIST and os.path.isdir(path):
                        pass
                    else:
                        raise
                outfile = None
                try:
                    full_filename = f'{path}/{filename}'
                    logger.debug(f"Try to write results to local file system: {full_filename}")
                    with open(full_filename, 'w') as outfile:
                        json.dump(results, outfile, indent=4)
                except IOError as e:
                    logger.error(f"Could not write results to file system: {e}")
                finally:
                    try:
                        if outfile:
                            outfile.close()
                    except:
                        pass # if outfile doesn't exist, no need to close it

                logger.info(f'Results for {len(results)} results are written to {full_filename}')
        elif not results:
            logger.info(colored("No findings found!", "green"))
            return_value = 0
        else:
            logger.info(f"Complete list with {len(results)} results is printed below:\n")
            click.echo(json.dumps(results, indent=4))

        return return_value

    def show_summary(self, results):
        '''
        Show summary containing test results.
        '''
        click.secho("\n\nSummary:\n", bold=True)
        headers = [
            'Source',
            'Action',
            'Resource',
            'Decision',
            'Policies',
        ]

        click.echo(tabulate(results, headers=headers, tablefmt="github"))
        click.echo("\n")




#! /usr/bin/env python3

"""
AWS IAM policies are notouriously complex, it is too easy to add some unintended
permissions and it is surprisingly difficult to identify these in heavily used AWS accounts.

Thankfully AWS has provided an IAM simulator that allows you to evaluate existing or
new policies for its behavior. Which is very nice, but doing this manually is quite
time consuming and it is unrealistic to test the entire environment for what you are
trying to do.

However, in good AWS spirit the simulator has an API and this tool provides automation
on top of it. It allows you to define the complete list of actions you want to evaluate
against what resources, which allows you to run these tests on a regular basis or (better)
integrate it in your CI/CD pipeline.
"""

# pylint: disable=broad-except,C0103,E0401,R0912,R0913,R0914,R0915,R1702,W0603,W1203

from __future__ import annotations
import json

import sys
import click

from aws_iam_tester.lib import AwsIamTester

from typing import Dict, List, Tuple, Optional, Any
from termcolor import colored
from outdated import check_outdated # type: ignore

from . import __version__

class DefaultGroup(click.Group):
    '''
    Allow a default command for a group
    '''
    ignore_unknown_options = True

    def __init__(self, *args, **kwargs):
        default_command = kwargs.pop('default_command', None)
        super(DefaultGroup, self).__init__(*args, **kwargs)
        self.default_cmd_name = None
        if default_command is not None:
            self.set_default_command(default_command)

    def set_default_command(self, command):
        if isinstance(command, str):
            cmd_name = command
        else:
            cmd_name = command.name
            self.add_command(command)
        self.default_cmd_name = cmd_name

    def parse_args(self, ctx, args):
        if not args and self.default_cmd_name is not None:
            args.insert(0, self.default_cmd_name)
        return super(DefaultGroup, self).parse_args(ctx, args)

    def get_command(self, ctx, cmd_name):
        if cmd_name not in self.commands and self.default_cmd_name is not None:
            ctx.args0 = cmd_name
            cmd_name = self.default_cmd_name
        return super(DefaultGroup, self).get_command(ctx, cmd_name)

    def resolve_command(self, ctx, args):
        cmd_name, cmd, args = super(DefaultGroup, self).resolve_command(ctx, args)
        args0 = getattr(ctx, 'args0', None)
        if args0 is not None:
            args.insert(0, args0)
        return cmd_name, cmd, args

@click.group(cls=DefaultGroup, default_command="account")
def cli():
    pass

@cli.command(name="account")
@click.option(
    '--number-of-runs', '-n',
    help='Run only a limited number of simulations, and then abort.',
    type=int,
    default=-1
    )
@click.option(
    '--dry-run', '-D',
    help='Dry run mode will not run the actual policy simulations. Default: False',
    is_flag=True,
    default=False
    )
@click.option(
    '--config-file', '-c',
    help='Config file location. Default: config.yml.',
    default='config.yml'
    )
@click.option(
    '--no-system-roles', '-N',
    help='Do not include non-user-assumable system roles.',
    is_flag=True,
    )
@click.option(
    '--write-to-file', '-w',
    help='Write results to file.',
    is_flag=True,
    )
@click.option(
    '--output-location', '-o',
    help='Output location, either s3 (start with s3://) or locally. Default: ./results',
    default='./results'
    )
@click.option(
    '--debug', '-d',
    help='Print debug messages.',
    is_flag=True,
    )
@click.version_option(version=__version__)
def check_aws_account(
        number_of_runs: int,
        dry_run: bool,
        config_file: str,
        no_system_roles: bool,
        write_to_file: bool,
        output_location: str,
        debug: bool
    ) -> int:
    """
    Checks an entire AWS account based on the provided configuration.

    Based on the findings the following return values will be generated:
    0: Upon successful completion with NO findings
    1: Upon successful completion with findings
    2: Upon failures
    """

    check_latest_version()

    try:
        tester = AwsIamTester(debug=debug)
        result = tester.check_account(
            number_of_runs=number_of_runs,
            dry_run=dry_run,
            config_file=config_file,
            no_system_roles=no_system_roles,
            write_to_file=write_to_file,
            output_location=output_location,
        )
        sys.exit(result)
    except Exception as e:
        click.echo(f"Exception occured: {e}")
        if debug:
            raise
        sys.exit(2)

@cli.command(name="access")
@click.option(
    '--user', '-u',
    help='User name that will be validated, if user and role is omitted all entities having access will be returned',
    default=None,
    )
@click.option(
    '--role', '-r',
    help='Role name that will be validated, if user and role is omitted all entities having access will be returned',
    default=None,
    )
@click.option(
    '--action', '-a',
    help='Action that will be validated',
    )
@click.option(
    '--resource', '-R',
    help="Resource that will be validated, default '*'",
    default="*"
    )
@click.option(
    '--json-output', '-j',
    help="Output in json format",
    is_flag=True,
    default=False
    )
@click.option(
    '--debug', '-d',
    help='Print debug messages.',
    is_flag=True,
    default=False
    )
@click.version_option(version=__version__)
def check_access(
        user: str,
        role: str,
        action: str,
        resource: str,
        json_output: bool,
        debug: bool,
    ):
    """
    Checks whether the provided IAM identity has permissions on the provided actions and resource.

    Based on the findings the following return values will be generated:
    0: Upon successful completion and allowed
    1: Upon successful completion and not allowed
    2: Upon failures
    """

    check_latest_version()

    try:
        tester = AwsIamTester(debug=debug)
        if user or role:
            allowed = tester.check_action(
                user=user,
                role=role,
                action=action,
                resource=resource,
                json_output=json_output,
            )
        else:
            allowed = tester.check_access(
                action=action,
                resource=resource,
                json_output=json_output,
            )
            sys.exit(0 if allowed else 1)
        sys.exit(0 if allowed else 1)
    except Exception as e:
        click.echo(f"Exception occured: {e}")
        if debug:
            raise
        sys.exit(2)

@cli.command(name="search")
@click.option(
    '--action', '-a',
    help='Action that will be validated',
    required=True,
    )
@click.option(
    '--resource', '-R',
    help="Resource that will be validated, default '*'",
    default="*"
    )
@click.option(
    '--json-output', '-j',
    help="Output in json format",
    is_flag=True,
    default=False
    )
@click.option(
    '--debug', '-d',
    help='Print debug messages.',
    is_flag=True,
    default=False
    )
@click.version_option(version=__version__)
def search_access(
        action: str,
        resource: str,
        json_output: bool,
        debug: bool,
    ):
    """
    Search which users and roles have access on the provided actions and resource.

    Based on the findings the following return values will be generated:
    0: Upon successful completion and allowed
    1: Upon successful completion and not allowed
    2: Upon failures
    """

    check_latest_version()

    try:
        tester = AwsIamTester(debug=debug)
        allowed = tester.check_access(
            action=action,
            resource=resource,
            json_output=json_output,
        )
        sys.exit(0 if allowed else 1)
    except Exception as e:
        click.echo(f"Exception occured: {e}")
        if debug:
            raise
        sys.exit(2)

def check_latest_version():
    # check for newer versions
    try:
        is_outdated, latest_version = check_outdated('aws-iam-tester', __version__)
        if is_outdated:
            click.echo(
                f'Your local version ({__version__}) is out of date! Latest is {latest_version}!'
            )
    except ValueError:
        # this happens when your local version is ahead of the pypi version,
        # which happens only in development
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aws_iam_tester', 'aws_iam_tester.lib']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.12.22,<2.0.0',
 'click>=7.1.1,<8.0.0',
 'outdated>=0.2.1,<0.3.0',
 'pyyaml>=5.3,<6.0',
 'tabulate>=0.8.7,<0.9.0',
 'termcolor>=1.1.0,<2.0.0']

entry_points = \
{'console_scripts': ['aws-iam-tester = aws_iam_tester.cli:cli']}

setup_kwargs = {
    'name': 'aws-iam-tester',
    'version': '0.3.6',
    'description': 'AWS IAM tester - simple command-line tool to check permissions handed out to IAM users and roles.',
    'long_description': '# Testing AWS IAM policies\n\n## Introduction\n\nAWS IAM policies are notouriously complex, it is too easy to add some unintended permissions and it is surprisingly difficult to identify these in heavily used AWS accounts.\n\nThankfully AWS has provided an [IAM simulator](https://docs.aws.amazon.com/IAM/latest/UserGuide/access_policies_testing-policies.html) that allows you to evaluate existing or new policies for its behavior. Which is very nice, but doing this manually is quite time consuming and it is unrealistic to test the entire environment for what you are trying to do.\n\nHowever, in good AWS spirit the simulator has an API and this tool provides automation on top of it. It allows you to define the complete list of actions you want to evaluate against what resources, which allows you to run these tests on a regular basis or (better) integrate it in your CI/CD pipeline.\n\n## Testing approach\n\nThe testing leverages AWS\' [IAM simulator (api)](https://docs.aws.amazon.com/IAM/latest/UserGuide/access_policies_testing-policies.html), that basically includes the same IAM evaluation logic that is applied when working in the console or using the cli. The beneits of this approach are:\n\n- It takes all different levels of policies into account. Think about permission boundaries, service control policies and so on.\n- It is an official service from AWS, so you can expect this to kept up to date over time.\n- The actual actions are evaluated, but NOT executed. Hence no need for cleaning up resources after testing.\n\n# Quick testing\n\nFor convenience, you can use this tool to quickly test whether a user has a specific permission on a particular resource:\n\n```bash\n$ aws-iam-tester access -u the_user -a \'glue:DeleteTable\'\nTest:\nSource:     arn:aws:iam::208912673223:user/the_user\nAction:     glue:DeleteTable\nResource:   *\nResult:     allowed\n\nMatched statements:\nPolicy:     admin_permissions\nType:       IAM Policy\nStart:      L3:C17\nEnd:        L8:C6\n```\n\nOr to see who has access to a particular resource and action by omitting both `user` and `role`:\n\n```bash\n$ aws-iam-tester access -a "s3:PutObject"\n..............\n\nSummary:\n\n| Source                                                   | Action       | Resource   | Decision   | Policies            |\n|----------------------------------------------------------|--------------|------------|------------|---------------------|\n| arn:aws:iam::XXXXXXXXXXXX:user/the_user                  | s3:PutObject | *          | allowed    | admins_permissions  |\n| arn:aws:iam::xxxxxxxxxxxx:role/the_role                  | s3:PutObject | *          | allowed    | AdministratorAccess |\n```\n\n## Finding out who has access to a particular action/resource combination\n\nIt might be useful to check your highly senstive resources and find out who exactly can access these. \n\n```bash\n$ aws-iam-tester access -a "s3:PutObject" -R "arn:aws:s3:::my-strictly-confidential-data"\n..............\n\nSummary:\n\n| Source                                                   | Action       | Resource                                   | Decision   | Policies            |\n|----------------------------------------------------------|--------------|--------------------------------------------|------------|---------------------|\n| arn:aws:iam::XXXXXXXXXXXX:user/the_user                  | s3:PutObject | arn:aws:s3:::my-strictly-confidential-data | allowed    | admins_permissions  |\n| arn:aws:iam::xxxxxxxxxxxx:role/the_role                  | s3:PutObject | arn:aws:s3:::my-strictly-confidential-data | allowed    | AdministratorAccess |\n```\n\n# Account testing\n\nHowever, the initial purpose of this tool is to check an entire account whether there are no users and/or roles having permissons which they should not have.\n\n## Configuration\n\nIn order to run, a configuration of the tests to run is required.\n\nA sample configuration (with only one test) is shown, in various steps.\n\nFirst there is a global section where you define settings which are applied to all tests (unless overruled, more on that later).\n\n```yaml\n---\nuser_landing_account: 0123456789 # ID of AWS Account that is allowed to assume roles in the test account\nglobal_exemptions: # The roles and/or users below will be ignored in all tests. Regular expressions are supported\n- "^arn:aws:iam::(\\\\d{12}):user/(.*)(ADMIN|admin)(.*)$"\n- "^arn:aws:iam::(\\\\d{12}):role/(.*)(ADMIN|admin)(.*)$"\n- "^arn:aws:iam::(\\\\d{12}):role/AWSCloudFormationStackSetExecutionRole$"\n```\n\nThen you define a list of tests, each consisting at least of a set of:\n- actions\n- resources\n- the expected result (should it fail or succeed)\n\n```yaml\n# List of tests to execute. In general the configurations follow the rules of the AWS IAM Policy Simulator.\n# For more information: https://docs.aws.amazon.com/IAM/latest/UserGuide/access_policies_testing-policies.html\ntests: \n- actions: # list of actions to validate\n  - "*:*"\n  - iam:*\n  - iam:AddUser*\n  - iam:Attach*\n  - iam:Create*\n  - iam:Delete*\n  - iam:Detach*\n  - iam:Pass*\n  - iam:Put*\n  - iam:Remove*\n  - iam:UpdateAccountPasswordPolicy\n  - sts:AssumeRole\n  - sts:AssumeRoleWithSAML\n  expected_result: fail # \'fail\' or \'succeed\'\n  resources: # list of resources to validate against\n  - "*"\n```\n\nRather than using all users and roles (without exemptions) you can also limit your test to a particular set of users and roles.\n\nThe test below does that, including defining a custom context that specifies multi factor authentication is disabled when running the test. By default the context under which the simulations are run assumes MFA is enabled, but you can override that with the `custom_context` element. For more information see the [AWS documentation](https://docs.aws.amazon.com/IAM/latest/UserGuide/access_policies_testing-policies.html).\n\n```yaml\n- actions: # Same list of actions, but now check (with a custom context) whether\n  - "*:*"\n  - iam:*\n  - iam:AddUser*\n  - iam:Attach*\n  - iam:Create*\n  - iam:Delete*\n  - iam:Detach*\n  - iam:Pass*\n  - iam:Put*\n  - iam:Remove*\n  - iam:UpdateAccountPasswordPolicy\n  - sts:AssumeRole\n  - sts:AssumeRoleWithSAML\n  expected_result: fail # \'fail\' or \'succeed\'\n  resources: # list of resources to validate against\n  - "*"\n  limit_to: # check this list for the admin users\n  - "^arn:aws:iam::(\\\\d*):user/(.*)(ADMIN|admin)(.*)$"\n  - "^arn:aws:iam::(\\\\d*):role/(.*)(ADMIN|admin)(.*)$"\n  # test if the admins are required to use multi factor authentication\n  custom_context: \n    - context_key_name: aws:MultiFactorAuthPresent\n      context_key_values: false\n      context_key_type: boolean\n```\n\nOr if you want to do that for **all** tests you can use the `global_limit_to`:\n\n```yaml\n---\nuser_landing_account: 0123456789 # ID of AWS Account that is allowed to assume roles in the test account\nglobal_limit_to: # These roles and/or users below will be u. Regular expressions are supported\n- "^arn:aws:iam::(\\\\d{12}):user/(.*)(ENGINEER|engineer)(.*)$"\n- "^arn:aws:iam::(\\\\d{12}):role/(.*)(SCIENTIST|scientist)(.*)$"\n```\n\nBelow an example where an additional set of roles is exempt from testing:\n\n```yaml\n- actions: # list of data centric actions\n  - redshift:GetClusterCredentials\n  - redshift:JoinGroup\n  - rds:Create*\n  - rds:Delete*\n  - rds:Modify*\n  - rds-db:connect\n  - s3:BypassGovernanceRetention\n  - s3:CreateBucket\n  - s3:DeleteBucket\n  - s3:DeleteBucketPolicy\n  - s3:PutBucketAcl\n  - s3:PutBucketPolicy\n  - s3:PutEncryptionConfiguration\n  - s3:ReplicateDelete\n  expected_result: fail # \'fail\' or \'succeed\'\n  resources: # list of resources to validate against\n  - "*"\n  exemptions: [\n  - "^arn:aws:iam::(\\\\d{12}):role/(.*)_worker$" # ignore this for the worker roles\n  ]\n```\n\nIf you want to run positive tests (i.e. tests that you need to succeed rather than fail), these `exemptions` don\'t work that well.\n\nIn that case you can limit your tests to a set of roles and users:\n\n```yaml\n- actions:\n  - s3:PutObject\n  expected_result: succeed\n  resources:\n  - "arn:aws:s3:::my_bucket/xyz/*"\n  limit_to: # if you specify this, test will only be performed for the sources below\n  - "^arn:aws:iam::(\\\\d{12}):role/my_worker$"\n```\n\n> Note that the exemptions are ignored when using a `limit_to` list.\n\n### Using a dynamic account id in the resource arn\n\nIn case you **need** to specify the `account id` in the resource arn, you can specificy this as follows:\n\n```yaml\n- actions:\n  - "secretsmanager:GetSecretValue"\n  expected_result: succeed\n  resources:\n  - "arn:aws:secretsmanager:eu-central-1:{account_id}:secret:my-secret/*"\n```\n\n### Using regular expressions for user and role matching\n\nRegular expressions can be used when checking the users and/or roles that need to be included in the tests. Hence they are supported for the following elements in the config file:\n\n- limit_to (the filtered list of users/roles to be used for a particular test)\n- global_limit_to (the filtered list of users/roles to be used for all tests)\n- exemptions (the filtered list of users/roles that should be excluded from a particular test)\n- global_exemptions (the filtered list of users/roles that should be excluded from all tests)\n\n> For all other elements, regular expression matching is NOT supported!\n\n## How to use\n\nAssuming you have define a config.yml in your local directory, then to run and write the outputs to the local `./results` directory:\n\n```bash\naws-iam-tester --write-to-file\n```\n\nUsing a specific config file:\n\n```bash\naws-iam-tester --config-file my-config.yml\n```\n\nUsing a specific output location:\n\n```bash\naws-iam-tester --output-location /tmp\n```\n\nOr write to s3:\n\n```bash\naws-iam-tester --output-location s3://my-bucket/my-prefix\n```\n\nInclude only roles that can be assumed by human beings:\n\n```bash\naws-iam-tester --no-include-system-roles\n```\n\n> Note: including system roles does NOT include the aws service roles.\n\nOr print debug output:\n\n```bash\naws-iam-tester --debug\n```\n\nTo run a limited number of evaluations (which helps speeding things up, and avoiding API throttling issues):\n\n```bash\naws-iam-tester --number-of-runs 10\n```\n\nFor more information, run `aws-iam-tester --help` for more instructions.\n\n## Return codes\n\nThe tester returns the following return codes:\n\n- 0 upon successful completion with NO findings\n- 1 upon successful completion with findings\n- 2 (or higher) on failures\n\n## Required permissions\n\nObviously the client has to run under an AWS security context that has sufficient permissions to query the IAM resources and run the simulator.\n\nThe following permissions are needed at the minimum:\n\n```yaml\n- sts:GetCallerIdentity\n- iam:ListAccountAliases\n- iam:ListRoles\n- iam:ListUsers\n- iam:SimulatePrincipalPolicy\n```\n\nAnd if you want to write the output to an s3 location, then obviously you need write access (`s3:PutObject`) to that particular location as well.\n\n## Unit testing\n\n`pytest` is being used for testing the various options.\n\nAs long as the `aws-iam-tester` module is installed, you can run the [tests](./tests).\n\nAfter installing `tox`, you can also simply run `$ tox`.',
    'author': 'Gerco Grandia',
    'author_email': 'gerco.grandia@4synergy.nl',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gercograndia/aws-iam-tester',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
